/**
 *  \file filter.c
 *  \author Nicola Ferri
 *  \brief Network filter component
 */

#include "proxy/filter.h"

#include <cmath>
#include <array>
#include <algorithm>
#include <stdexcept>

#include <boost/log/trivial.hpp>
#include <boost/bind.hpp>

#include "util/rng.h"

namespace selink { namespace proxy {

        void filter::start()
        {
            /* Choose which socket we need to connect, depending on the side.
               A is always the local endpoint, and B is always the remote endpoint,
               therefore fb is encrypted (possibly, in encrypt mode), fb is not.
               However, if we are bridging for the server, we must connect to the local
               endpoint, which is accepting connections; if we are bridging for the client,
               we must connect to the remote endpoint, the one for which the
               redirected connection was intended.
            */

			if (redirect_info.side == side_connect) {
				fa_connected = true;
			}
			else {
				fb_connected = true;
			}

            boost::asio::ip::tcp::socket& conn_socket =
                (redirect_info.side == side_connect) ?
                (fb) : (fa);

            if (redirect_info.encrypt && (ep_.status() == selink::provider::wait_config)) {
                // encrypt mode but provider not ready
                BOOST_LOG_TRIVIAL(error) << "Encryption provider not ready (wait_config)";
				stop();
                return;
            }

            ep_ref = ep_.ref;

            // create endpoint from redirect info
            boost::asio::ip::address addr;
            if (redirect_info.host_ip.empty()) {
                addr = boost::asio::ip::address::from_string(redirect_info.host);
            }
            else if (redirect_info.host_ip.size() == 4) {
                std::array<uint8_t, 4> tmp;
                std::copy(redirect_info.host_ip.begin(), redirect_info.host_ip.end(), tmp.begin());
                addr = boost::asio::ip::address_v4(tmp);
            }
            else if (redirect_info.host_ip.size() == 16) {
                std::array<uint8_t, 16> tmp;
                std::copy(redirect_info.host_ip.begin(), redirect_info.host_ip.end(), tmp.begin());
                addr = boost::asio::ip::address_v6(tmp);
            }
            else {
                BOOST_LOG_TRIVIAL(error) << "Redirect information not valid";
                stop();
                return;
            }

			redirect_info.asio_endpoint = tcp::endpoint(addr, redirect_info.port);

            if (redirect_info.side != side_connect && redirect_info.side != side_accept) {
                // this should not happen
                stop();
                return;
            }


            if (redirect_info.encrypt) {
                // allocate buffers needed for encryption
                eb = std::unique_ptr<encryption_buffers>(new encryption_buffers());
                
                // follow encrypt mode path
				if (redirect_info.side == side_connect) {
					entry_svc_incoming_a();
				}
				else {
					entry_gw_incoming_b();
				}
            }
            else {
                // follow bypass mode path
                conn_socket.async_connect(
					redirect_info.asio_endpoint,
                    boost::bind(
                        &filter::handle_connect_bypass,
                        shared_from_this(),
                        boost::asio::placeholders::error));
            }
        }

        tcp::socket& filter::get_fa()
        {
            return fa;
        }
        tcp::socket& filter::get_fb()
        {
            return fb;
        }

        filter::pointer filter::create(boost::asio::io_service& io, boost::asio::strand& strand_encrypt, provider::provider& ep)
        {
            return filter::pointer(new filter(io, strand_encrypt, ep));
        }

        filter::filter(boost::asio::io_service& io, boost::asio::strand& strand_encrypt, provider::provider& ep)
            : fa(io), fb(io), ep_(ep), eb(nullptr), strand_encrypt_(strand_encrypt),
            buf_ab_raw(1024), buf_ba_raw(1024)
        { }

        void filter::set_redirect_info(std::string host,
            uint16_t port, int side, bool encrypt)
        {
            redirect_info.host_ip.clear();
            redirect_info.host = host;
            redirect_info.port = port;
            redirect_info.side = side;
            redirect_info.encrypt = encrypt;
			redirect_info.key_id = -1;
        }

        void filter::set_redirect_info(std::vector<uint8_t> host,
            uint16_t port, int side, bool encrypt)
        {
            redirect_info.host.clear();
            redirect_info.host_ip = host;
            redirect_info.port = port;
            redirect_info.side = side;
            redirect_info.encrypt = encrypt;
            redirect_info.key_id = -1;
        }






		// -----

		// svc
		void filter::entry_svc_incoming_a()
		{
			// connect to B
			try {
				fb.async_connect(
					redirect_info.asio_endpoint,
					strand_encrypt_.wrap(boost::bind(
						&filter::handle_svc_connected_to_b,
						shared_from_this(),
						boost::asio::placeholders::error)));
			}
			catch (std::exception&) {
				stop();
			}
		}

		void filter::handle_svc_connected_to_b(const boost::system::error_code& error)
		{
			// send D1 to B
			if (error) {
				BOOST_LOG_TRIVIAL(error) << "Error Connecting to B";
				stop();
				return;
			}
			fb_connected = true;
			
			selink::util::rng::rand_bytes(eb->link.salt, SELINK_SALT_SIZE);
			
			bool b = ep_.import_keys_and_algorithms(&(eb->link));
			if (!b) {
				BOOST_LOG_TRIVIAL(error) << "Error importing keys and algorithms";
				stop();
				return;
			}
			buf_ensure(buf_ab_raw, SELINK_D1_SIZE_MAX);
			size_t request_len = 0;

			SElink_raw_d1_write(&(eb->link), &request_len, &buf_ab_raw[0]);
			buf_ab_raw.resize(request_len);

			fb.async_send(
				boost::asio::buffer(buf_ab_raw, buf_ab_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_svc_d1_sent_to_b,
					shared_from_this(),
					boost::asio::placeholders::error)));

		}

		void filter::handle_svc_d1_sent_to_b(const boost::system::error_code& error)
		{
			if (error) {
				BOOST_LOG_TRIVIAL(error) << "Error sending D1 to B";
				stop();
				return;
			}
			// receive D2 from B
			buf_ensure(buf_ab_raw, SELINK_D2_SIZE);
			boost::asio::async_read(fb,
				boost::asio::buffer(buf_ab_raw, buf_ab_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_svc_d2_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));
		}

		void filter::handle_svc_d2_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error) {
				BOOST_LOG_TRIVIAL(error) << "Error receiving D2 from B";
				stop();
				return;
			}

			// decode D2
			bool b = SElink_raw_d2_read(&(eb->link), buf_ab_raw.size(), &buf_ab_raw[0]);
			if (!b) {
				BOOST_LOG_TRIVIAL(error) << "Error decoding D2";
				stop();
				return;
			}
			
			// check if negotiation succeeded
			if (eb->link.choice.algorithm_index == -1 || eb->link.choice.key_index == -1) {
				stop();
				return;
			}

			redirect_info.key_id = eb->link.choice.key.id;
			redirect_info.algo_id = eb->link.choice.algorithm.id;
			
			if (!create_sessions()) {
				stop();
				return;
			}

			// receive DATA from A
			buf_ensure(buf_ab_raw, 4096);
			buf_ab_raw.resize((std::min)(buf_ab_raw.capacity(), (size_t)(SELINK_DATA_MAX + SELINK_H_SIZE)));
			fa.async_read_some(
				boost::asio::buffer(&buf_ab_raw[SELINK_H_SIZE], buf_ab_raw.size() - SELINK_H_SIZE),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_data_received_from_a,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

			// receive HDR from B
			buf_ensure(buf_ba_raw, SELINK_HEADER_REQUIRED);
			boost::asio::async_read(fb,
				boost::asio::buffer(buf_ba_raw, buf_ba_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_hdr_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

		}

		// gw
		void filter::entry_gw_incoming_b()
		{
			buf_ensure(buf_ab_raw, SELINK_D1_REQUIRED);
			
			// receive D1 header from B
			boost::asio::async_read(fb,
				boost::asio::buffer(buf_ab_raw, buf_ab_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_gw_d1h_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

		}
		void filter::handle_gw_d1h_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error || bytes_transferred != SELINK_D1_REQUIRED) {
				stop();
				return;
			}
			size_t size = 0;

			bool b = SElink_raw_d1_get_size(&(eb->link), buf_ab_raw.size(), &buf_ab_raw[0], &size);
			if (!b) {
				stop();
				return;
			}

			buf_ensure(buf_ab_raw, size);
			// receive D1 from B
			boost::asio::async_read(fb,
				boost::asio::buffer(&buf_ab_raw[SELINK_D1_REQUIRED], buf_ab_raw.size() - SELINK_D1_REQUIRED),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_gw_d1_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

		}
		void filter::handle_gw_d1_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error || bytes_transferred != buf_ab_raw.size() - SELINK_D1_REQUIRED) {
				stop();
				return;
			}
			if (!SElink_raw_d1_read(&(eb->link), buf_ab_raw.size(), &buf_ab_raw[0])) {
				stop();
				return;
			}

			if (!ep_.import_keys_and_algorithms(&(eb->link))) {
				stop();
				return;
			}

			if (!SElink_raw_d1_match(&(eb->link), buf_ab_raw.size(), &buf_ab_raw[0])) {
				BOOST_LOG_TRIVIAL(error) << "Error could not find matching key/algorithm";
				stop();
				return;
			}

			redirect_info.key_id = eb->link.choice.key.id;
			redirect_info.algo_id = eb->link.choice.algorithm.id;

			size_t size = 0;
			buf_ensure(buf_ab_raw, SELINK_D2_SIZE);
			SElink_raw_d2_write(&(eb->link), &size, &buf_ab_raw[0]);
			buf_ab_raw.resize(size);

			// send D2 to B
			boost::asio::async_write(fb,
				boost::asio::buffer(buf_ab_raw, buf_ab_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_gw_d2_sent_to_b,
					shared_from_this(),
					boost::asio::placeholders::error)));

		}
		void filter::handle_gw_d2_sent_to_b(const boost::system::error_code& error)
		{
			if (error) {
				stop();
				return;
			}
			
			if (!create_sessions()) {
				stop();
				return;
			}

			// connect to A
			try {
				fa.async_connect(
					redirect_info.asio_endpoint,
					strand_encrypt_.wrap(boost::bind(
						&filter::handle_gw_connected_to_a,
						shared_from_this(),
						boost::asio::placeholders::error)));
			}
			catch (std::exception&) {
				stop();
				return;
			}
		}
		void filter::handle_gw_connected_to_a(const boost::system::error_code& error)
		{
			if (error) {
				stop();
				return;
			}
			fa_connected = true;
			
			// receive DATA from A
			buf_ensure(buf_ab_raw, 4096);
			buf_ab_raw.resize((std::min)(buf_ab_raw.capacity(), (size_t)(SELINK_DATA_MAX + SELINK_H_SIZE)));
			fa.async_read_some(
				boost::asio::buffer(&buf_ab_raw[SELINK_H_SIZE], buf_ab_raw.size() - SELINK_H_SIZE),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_data_received_from_a,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

			// receive HDR from B
			buf_ensure(buf_ba_raw, SELINK_HEADER_REQUIRED);
			boost::asio::async_read(fb,
				boost::asio::buffer(buf_ba_raw, buf_ba_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_hdr_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));
		}

		// data
		void filter::handle_data_received_from_a(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error) {
				stop();
				return;
			}

			bool expand_buffer = (buf_ab_raw.size() == (bytes_transferred + SELINK_H_SIZE)) && (bytes_transferred < SELINK_DATA_MAX);

			size_t padded_size = SELINK_H_SIZE + bytes_transferred;
			if (0 != padded_size % SELINK_BLOCK_SIZE) {
				padded_size += 16 - (padded_size % SELINK_BLOCK_SIZE);
			}

			size_t packet_size = SElink_raw_packet_size(bytes_transferred);

			// encrypt data, send HDR+ENCDATA to B
			
			buf_ensure(eb->buf_ab_enc, 
				packet_size - SELINK_AUTH_SIZE + SELINK_AUTH_MAX);

			selink::util::rng::rand_bytes(&(eb->buf_ab_enc[0]), SELINK_BLOCK_SIZE);
			
			buf_ensure(buf_ab_raw, padded_size);
			if (padded_size != SELINK_H_SIZE + bytes_transferred) {
				memset(&buf_ab_raw[SELINK_H_SIZE + bytes_transferred], 0,
					padded_size - (SELINK_H_SIZE + bytes_transferred));
			}

			SElink_raw_h_write(&(eb->link), &(buf_ab_raw[0]), bytes_transferred);

			auto e = session_ab->update(
				true, &(eb->buf_ab_enc[0]), 
				true, padded_size / SELINK_BLOCK_SIZE, &(buf_ab_raw[0]), &(eb->buf_ab_enc[SELINK_BLOCK_SIZE]),
				true);
			if (e != provider::session::error::ok) {
				eb->packet_decrypt_state.error = eb->packet_decrypt_state.error_provider;
				stop();
				return;
			}

			eb->buf_ab_enc.resize(packet_size);

			if (expand_buffer) {
				buf_ensure(buf_ab_raw, buf_ab_raw.capacity() + 1);
			}

			boost::asio::async_write(fb,
				boost::asio::buffer(eb->buf_ab_enc, eb->buf_ab_enc.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_hdr_encdata_sent_to_b,
					shared_from_this(),
					boost::asio::placeholders::error)));
			
		}

		void filter::handle_hdr_encdata_sent_to_b(const boost::system::error_code& error)
		{
			if (error) {
				stop();
				return;
			}

			// receive DATA from A
			buf_ab_raw.resize((std::min)(buf_ab_raw.capacity(), (size_t)(SELINK_DATA_MAX + SELINK_H_SIZE)));
			fa.async_read_some(
				boost::asio::buffer(&buf_ab_raw[SELINK_H_SIZE], buf_ab_raw.size()-SELINK_H_SIZE),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_data_received_from_a,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

		}
		void filter::handle_hdr_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error || (bytes_transferred != SELINK_HEADER_REQUIRED)) {
				stop();
				return;
			}

			// decrypt HDR
			size_t bytes_to_decrypt = SELINK_HEADER_REQUIRED - SELINK_BLOCK_SIZE;
			buf_ensure(eb->buf_ba_dec, SELINK_HEADER_REQUIRED - SELINK_BLOCK_SIZE);
			auto e = session_ba->update(
				true, &buf_ba_raw[0],
				true, bytes_to_decrypt / SELINK_BLOCK_SIZE, &buf_ba_raw[SELINK_BLOCK_SIZE], &(eb->buf_ba_dec[0]),
				false);
			if (e != provider::session::error::ok) {
				eb->packet_decrypt_state.error = eb->packet_decrypt_state.error_provider;
				stop();
				return;
			}
			
			size_t data_size = 0;
			SElink_raw_h_read(&(eb->link), &(eb->buf_ba_dec[0]), &data_size);
			if (data_size == 0 || data_size > SELINK_DATA_MAX) {
				eb->packet_decrypt_state.error = eb->packet_decrypt_state.error_header;
				stop();
				return;
			}

			eb->packet_decrypt_state.bytes_decrypted = bytes_to_decrypt;
			eb->packet_decrypt_state.message_size = data_size;
			eb->packet_decrypt_state.packet_size = SElink_raw_packet_size(data_size);
			eb->packet_decrypt_state.padded_size = SELINK_H_SIZE + data_size;
			if (0 != (eb->packet_decrypt_state.padded_size % SELINK_BLOCK_SIZE)) {
				eb->packet_decrypt_state.padded_size += SELINK_BLOCK_SIZE - (eb->packet_decrypt_state.padded_size % SELINK_BLOCK_SIZE);
			}
			eb->buf_ba_needed = eb->packet_decrypt_state.packet_size - bytes_transferred;

			buf_ensure(buf_ba_raw, eb->buf_ba_needed);

			// receive ENCDATA from B
			boost::asio::async_read(fb,
				boost::asio::buffer(buf_ba_raw, buf_ba_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_encdata_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));

		}
		void filter::handle_encdata_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred)
		{
			if (error) {
				stop();
				return;
			}

			// decrypt ENCDATA
			
			buf_ensure(eb->buf_ba_dec, eb->packet_decrypt_state.padded_size + SELINK_AUTH_MAX);
			size_t bytes_to_decrypt = eb->packet_decrypt_state.padded_size - eb->packet_decrypt_state.bytes_decrypted;
			auto e = session_ba->update(
				false, NULL,
				true, bytes_to_decrypt / SELINK_BLOCK_SIZE, &buf_ba_raw[0],
				&(eb->buf_ba_dec[eb->packet_decrypt_state.bytes_decrypted]),
				true);
			if (e != provider::session::error::ok) {
				eb->packet_decrypt_state.error = eb->packet_decrypt_state.error_provider;
				stop();
				return;
			}
			
			if (memcmp(&eb->buf_ba_dec[eb->packet_decrypt_state.padded_size], &buf_ba_raw[bytes_to_decrypt], SELINK_AUTH_SIZE)) {
				eb->packet_decrypt_state.error = eb->packet_decrypt_state.error_auth;
				stop();
				return;
			}

			// send DATA to A
			boost::asio::async_write(fa,
				boost::asio::buffer(&(eb->buf_ba_dec[SELINK_H_SIZE]), eb->packet_decrypt_state.message_size),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_data_sent_to_a,
					shared_from_this(),
					boost::asio::placeholders::error)));
		}
		void filter::handle_data_sent_to_a(const boost::system::error_code& error)
		{
			if (error) {
				stop();
				return;
			}

			// receive HDR from B
			buf_ba_raw.resize(SELINK_HEADER_REQUIRED);
			boost::asio::async_read(fb, 
				boost::asio::buffer(buf_ba_raw, buf_ba_raw.size()),
				strand_encrypt_.wrap(boost::bind(
					&filter::handle_hdr_received_from_b,
					shared_from_this(),
					boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred)));
		}

		// -----











		/*


        void filter::handle_connect_encrypt(const boost::system::error_code& error) {
            if (error || ep_.ref != ep_ref) {
                stop();
            }
            else {
                // create encrypting session A->B
                selink::util::rng::rand_bytes(eb->nonce_ab, provider::provider::nonce_size);
                session_ab = ep_.create_session(
                    redirect_info.key_id, eb->nonce_ab, provider::session::direction::encrypt);

                // send nonce to B
                boost::asio::async_write(fb,
                    boost::asio::buffer(eb->nonce_ab, provider::provider::nonce_size),
                    strand_encrypt_.wrap(
                        boost::bind(&filter::handle_fb_sent_nonce,
                        shared_from_this(),
                        boost::asio::placeholders::error)));


                // receive nonce from B
                boost::asio::async_read(fb,
                    boost::asio::buffer(eb->nonce_ba, provider::provider::nonce_size),
                    strand_encrypt_.wrap(
                        boost::bind(
                            &filter::handle_fb_read_nonce,
                            shared_from_this(),
                            boost::asio::placeholders::error,
                            boost::asio::placeholders::bytes_transferred)));
            }
        }


        // A <-> F
        void filter::handle_fa_sent_data(
            const boost::system::error_code& error)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }

            // read iv + header from B
            boost::asio::async_read(fb,
                boost::asio::buffer(buf_ba_raw, eb->bytes_needed_for_header),
                strand_encrypt_.wrap(
                    boost::bind(
                        &filter::handle_fb_read_hdr,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
        }

        void filter::handle_fa_read_data(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            // if the buffer was filled entirely, increase its size
            if ((bytes_transferred == buf_ab_raw.size() - encryption_overhead) &&
                (buf_ab_raw.size() < buffer_max))
            {
                buf_ensure(buf_ab_raw, buf_ab_raw.size() + 1);
            }
            buf_ensure(eb->buf_ab_enc, buf_ab_raw.size());

            // encrypt
            int enc_size = packet_encrypt(bytes_transferred, &buf_ab_raw[0], &(eb->buf_ab_enc[0]));
            if (enc_size < 1) {
                stop();
                return;
            }
            // send data to B
            boost::asio::async_write(fb,
                boost::asio::buffer(eb->buf_ab_enc, enc_size),
                strand_encrypt_.wrap(
                    boost::bind(&filter::handle_fb_sent_data,
                        shared_from_this(),
                        boost::asio::placeholders::error)));
        }

        // F <-> B
        void filter::handle_fb_sent_nonce(
            const boost::system::error_code& error)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            // read some data from A
            fa.async_read_some(
                boost::asio::buffer(
                    &buf_ab_raw[header_size], buf_ab_raw.size() - encryption_overhead),
                strand_encrypt_.wrap(
                    boost::bind(
                        &filter::handle_fa_read_data,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
        }
        void filter::handle_fb_sent_data(
            const boost::system::error_code& error)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            // read some data from A
            fa.async_read_some(
                boost::asio::buffer(
                    &buf_ab_raw[header_size], buf_ab_raw.size() - encryption_overhead),
                strand_encrypt_.wrap(
                    boost::bind(
                        &filter::handle_fa_read_data,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
        }

        void filter::handle_fb_read_nonce(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            if (bytes_transferred != provider::provider::nonce_size) {
                stop();
                return;
            }
            // create decrypting session B->A
            session_ba = ep_.create_session(
                redirect_info.key_id, eb->nonce_ba, provider::session::direction::decrypt);

            // read iv + header from B
            boost::asio::async_read(fb,
                boost::asio::buffer(buf_ba_raw, eb->bytes_needed_for_header),
                strand_encrypt_.wrap(
                    boost::bind(
                        &filter::handle_fb_read_hdr,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
        }

        void filter::handle_fb_read_hdr(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            if (bytes_transferred != eb->bytes_needed_for_header) {
                stop();
                return;
            }

            // decrypt header
            buf_ensure(eb->buf_ba_dec, buf_ba_raw.size());
            packet_decrypt_hdr(&buf_ba_raw[0], &(eb->buf_ba_dec[0]));
            if (eb->packet_decrypt_state.error != eb->packet_decrypt_state.error_none) {
                stop();
                return;
            }

            // compute bytes needed to complete the packet
            eb->buf_ba_needed = eb->packet_decrypt_state.packet_size - eb->bytes_needed_for_header;
            // encryption::session::padded_size(message_size) + encryption::session::auth_size;

            // receive packet message+auth
            buf_ensure(buf_ba_raw, eb->packet_decrypt_state.packet_size);
            boost::asio::async_read(fb,
                boost::asio::buffer(&buf_ba_raw[eb->bytes_needed_for_header], eb->buf_ba_needed),
                strand_encrypt_.wrap(
                    boost::bind(
                        &filter::handle_fb_read_data,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred)));
        }

        void filter::handle_fb_read_data(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error || ep_.ref != ep_ref) {
                stop();
                return;
            }
            if (bytes_transferred != eb->buf_ba_needed) {
                stop();
                return;
            }

            // decrypt
            buf_ensure(eb->buf_ba_dec, buf_ba_raw.size());
            packet_decrypt_msg(&buf_ba_raw[0], &(eb->buf_ba_dec[0]));
            if (eb->packet_decrypt_state.error != eb->packet_decrypt_state.error_none) {
                stop();
                return;
            }

            // send data to A
            boost::asio::async_write(fa,
                boost::asio::buffer(
                    &(eb->buf_ba_dec[header_size]), eb->packet_decrypt_state.message_size),
                strand_encrypt_.wrap(
                    boost::bind(&filter::handle_fa_sent_data,
                        shared_from_this(),
                        boost::asio::placeholders::error)));
        }
		*/

        // --- Bypass ---

        void filter::handle_connect_bypass(const boost::system::error_code& error)
        {
            if (error) {
                BOOST_LOG_TRIVIAL(trace) << "handle_connect_bypass error " << error.value();
                stop();
                return;
            }
			fa_connected = true;
			fb_connected = true;

            // start reading from A and B
            fa.async_read_some(
                boost::asio::buffer(buf_ab_raw),
                boost::bind(
                    &filter::handle_fa_read_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred));
            fb.async_read_some(
                boost::asio::buffer(buf_ba_raw),
                boost::bind(
                    &filter::handle_fb_read_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred));
        }

        void filter::handle_fa_sent_bypass(
            const boost::system::error_code& error)
        {
            if (error) {
                BOOST_LOG_TRIVIAL(trace) << "handle_fa_sent_bypass error " << error.value();
                stop();
                return;
            }

            fb.async_read_some(
                boost::asio::buffer(buf_ba_raw),
                boost::bind(
                    &filter::handle_fb_read_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred));
        }
        void filter::handle_fb_sent_bypass(
            const boost::system::error_code& error)
        {
            if (error) {
                BOOST_LOG_TRIVIAL(trace) << "handle_fb_sent_bypass error " << error.value();
                stop();
                return;
            }

            fa.async_read_some(
                boost::asio::buffer(buf_ab_raw),
                boost::bind(
                    &filter::handle_fa_read_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred));
        }
        void filter::handle_fa_read_bypass(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error) {
                BOOST_LOG_TRIVIAL(trace) << "handle_fa_read_bypass error " << error.value();
                stop();
                return;
            }
            // if the buffer was filled entirely, increase its size
            if (bytes_transferred == buf_ab_raw.size() &&
                (buf_ab_raw.size() < buffer_max))
            {
                buf_ensure(buf_ab_raw, buf_ab_raw.size() + 1);
            }
            // send data to B
            boost::asio::async_write(fb,
                boost::asio::buffer(buf_ab_raw, bytes_transferred),
                boost::bind(&filter::handle_fb_sent_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error));
        }
        void filter::handle_fb_read_bypass(
            const boost::system::error_code& error,
            std::size_t bytes_transferred)
        {
            if (error) {
                BOOST_LOG_TRIVIAL(trace) << "handle_fb_read_bypass error " << error.value();
                stop();
                return;
            }
            // if the buffer was filled entirely, increase its size
            if (bytes_transferred == buf_ba_raw.size() &&
                (buf_ba_raw.size() < buffer_max))
            {
                buf_ensure(buf_ba_raw, buf_ba_raw.size() + 1);
            }
            // send data to A
            boost::asio::async_write(fa,
                boost::asio::buffer(buf_ba_raw, bytes_transferred),
                boost::bind(&filter::handle_fa_sent_bypass,
                    shared_from_this(),
                    boost::asio::placeholders::error));
        }


        void filter::stop()
        {
            boost::mutex::scoped_lock lock(mut);
            try {
                if (fa_connected && fa.is_open()) {
                    fa.shutdown(tcp::socket::shutdown_both);
                    fa.close();
                }
            }
            catch (std::exception&) {}
            try {
                if (fb_connected && fb.is_open()) {
                    fb.shutdown(tcp::socket::shutdown_both);
                    fb.close();
                }
            }
            catch (std::exception&) {}
        }

		bool filter::create_sessions() {
			// #TODO algo id
			session_ab = ep_.create_session(redirect_info.key_id, redirect_info.algo_id, eb->link.salt,
				provider::session::direction::encrypt);
			if (!session_ab) {
				BOOST_LOG_TRIVIAL(error) << "Error creating session AB";
				return false;
			}
			session_ba = ep_.create_session(redirect_info.key_id, redirect_info.algo_id, eb->link.salt,
				provider::session::direction::decrypt);
			if (!session_ba) {
				BOOST_LOG_TRIVIAL(error) << "Error creating session BA";
				session_ab.reset();
				return false;
			}
			return true;
		}

        void filter::buf_ensure(std::vector<uint8_t>& vec, size_t needed) {
            size_t next = (size_t)pow(2, ceil(log(needed) / log(2)));
			if(vec.capacity() < next)
				vec.reserve(next);
			vec.resize(needed);
        }

} }  // namespace selink::proxy
