/**
 *  \file filter.h
 *  \author Nicola Ferri
 *  \brief Network filter component
 *
 *  This component bridges two sockets, encrypting or decrypting traffic as needed.
 */

#pragma once
#include "util/os.h"

#include <cstdint>
#include <memory>
#include <string>
#include <vector>

#include <boost/thread.hpp>
#include <boost/asio.hpp>
#include <boost/enable_shared_from_this.hpp>

#include "provider/provider.h"

namespace selink { namespace proxy {

    using boost::asio::ip::tcp;

    /** This class implements the logic for communication filtering.
     *  It takes two sockets, FA (filter <-> A) and FB (filter <-> B)
     *  and bridges the unencrypted connection FA to the encrypted connection FB.
     *  Depending on the configuration, it may also proxy the connection
     *  as is (bypass mode)
     */
    class filter : public boost::enable_shared_from_this<filter> {
    public:
        /** Protocol constants */
        enum {
            buffer_max = (64 * 1024),
            header_size = 4,
            encryption_overhead = (
                header_size + provider::session::iv_size +
                provider::session::iv_size),
            message_max = (buffer_max - encryption_overhead)
        };
        
        enum {
            /** server side proxy */
            side_accept,
            /** client side proxy */
            side_connect
        };
        typedef boost::shared_ptr<filter> pointer;

        /** \brief start the bridging.
         *
         *  The filter will connect to the other endpoint and filter
         *  any communication
         */
        void start();

        /** \brief Get socket FA */
        tcp::socket& get_fa();

        /** \brief Get socket FB */
        tcp::socket& get_fb();

        /** \brief Create a new filter
         *  \param io boost.asio io service
         *  \param strand_encrypt common strand for all encrypted connections
         *  \param ep encryption provider
         *  
         *  If the encryption provider does not support concurrent operations,
         *  it is important to always specify the same strand, to ensure mutual exclusion
         *  of the operations using the provider.
         */
        static pointer create(
            boost::asio::io_service& io, boost::asio::strand& strand_encrypt, provider::provider& ep);

        /** \brief Set parameters for redirection
         *  \param host the host to which the connection will be redirected
         *  \param port the port to which the connection will be redirected
         *  \param encrypt if true, the connection will be encrypted
         *  \param key_id key to be used to encrypt the connection
         */
        void set_redirect_info(std::string host,
            uint16_t port, int side, bool encrypt);

        /** \brief Set parameters for redirection
        *  \param host ip address of the host to which the connection will be redirected
        *  \param port the port to which the connection will be redirected
        *  \param encrypt if true, the connection will be encrypted
        *  \param key_id key to be used to encrypt the connection
        */
        void set_redirect_info(std::vector<uint8_t> host,
            uint16_t port, int side, bool encrypt);

    private:

        /** Redirection parameters needed to setup the bridge */  
        struct redirect_info_t {
            int side;
            std::string host;
            std::vector<uint8_t> host_ip;
            uint16_t port{ 0 };
            bool encrypt{ false };
            int key_id{ -1 };
			int algo_id{ -1 };
			boost::asio::ip::tcp::endpoint asio_endpoint;
        } redirect_info;

        tcp::socket fa;
        tcp::socket fb;

		bool fa_connected{ false };
		bool fb_connected{ false };


        provider::provider& ep_;
		bool create_sessions();
        unsigned int ep_ref{ 0 };

        boost::mutex mut;

        /** See create(). */
        filter(boost::asio::io_service& io, boost::asio::strand& strand_encrypt, provider::provider& ep);


        /** @{
         *  Handlers for encryption mode path
         */

		// svc
		void entry_svc_incoming_a();
		void handle_svc_connected_to_b(const boost::system::error_code& error);
		void handle_svc_d1_sent_to_b(const boost::system::error_code& error);
		void handle_svc_d2_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred);

		// gw
		void entry_gw_incoming_b();
		void handle_gw_d1h_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred);
		void handle_gw_d1_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred);
		void handle_gw_d2_sent_to_b(const boost::system::error_code& error);
		void handle_gw_connected_to_a(const boost::system::error_code& error);

		// data
		void handle_data_received_from_a(const boost::system::error_code& error,
			std::size_t bytes_transferred);
		void handle_hdr_encdata_sent_to_b(const boost::system::error_code& error);
		void handle_hdr_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred);
		void handle_encdata_received_from_b(const boost::system::error_code& error,
			std::size_t bytes_transferred);
		void handle_data_sent_to_a(const boost::system::error_code& error);

        /** @} */


        /** @{
        *  Handlers for bypass mode path
        */
        void handle_connect_bypass(const boost::system::error_code& error);
        void handle_fa_sent_bypass(const boost::system::error_code& error);
        void handle_fa_read_bypass(const boost::system::error_code& error,
            std::size_t bytes_transferred);
        void handle_fb_sent_bypass(const boost::system::error_code& error);
        void handle_fb_read_bypass(const boost::system::error_code& error,
            std::size_t bytes_transferred);
        /** @} */

        /** Structures for encryption mode */
        struct encryption_buffers {
            struct {
                enum { error_none = 0, error_auth, error_provider, error_header, error_init };
				int error{ error_init };
                size_t bytes_decrypted{ 0 };
                size_t message_size{ 0 };
                size_t packet_size{ 0 };
				size_t padded_size{ 0 };
            } packet_decrypt_state;

            std::vector<uint8_t> buf_ab_enc;
            std::vector<uint8_t> buf_ba_dec;
            size_t buf_ba_needed;
            size_t buf_ba_message_size;

			SElink_raw link;

            encryption_buffers() :
                buf_ab_enc(1024), buf_ba_dec(1024),
                buf_ba_needed(0), buf_ba_message_size(0)
            {
				SElink_raw_init(&link);
            }
        };

        std::unique_ptr<encryption_buffers> eb{ nullptr };
        boost::asio::strand& strand_encrypt_;
        
        /** Session A-to-B */
        std::shared_ptr<provider::session> session_ab;

        /** Session B-to-A */
        std::shared_ptr<provider::session> session_ba;

        /** Raw data buffers */
        std::vector<uint8_t> buf_ab_raw;
        std::vector<uint8_t> buf_ba_raw;

        /** \brief Close both sockets and terminate bridging */
        void stop();

        /** \brief Ensure buffer capacity
         *  \param vec vector of uint8_t, its size must be a power of 2
         *  \param needed bytes needed in buffer
         *  
         *  the buffer will be resized to a capacity greater than or equal to 'needed'.
         *  Increments are in powers of 2.
         */
        static void buf_ensure(std::vector<uint8_t>& vec, size_t needed);
    };

} }  // namespace selink::proxy




