#include "SElink.h"

static uint8_t selink_magic[2] = { 0x61, 0xde };
uint8_t selink_ver = 1;

static uint16_t _SElink_crypto_init(SElink* ctx);
static bool _SElink_raw_check_packet(const uint8_t* data, uint8_t type);


enum {
	SELINK_TYPE_D1 = 0xD1,
	SELINK_TYPE_D2 = 0xD2,
	SELINK_TYPE_S1 = 0x51
};


void SElink_raw_init(SElink_raw* sr)
{
	sr->neg.algorithms = NULL;
	sr->neg.algorithms_cap = 0;
	sr->neg.algorithms_len = 0;
	sr->neg.keys = NULL;
	sr->neg.keys_cap = 0;
	sr->neg.keys_len = 0;

	memset(sr->conf.additional, 0, SELINK_ADD_SIZE);

	sr->choice.key.id = SE3_KEY_INVALID;
	sr->choice.algorithm.id = SE3_ALGO_INVALID;
}

void SElink_raw_reserve_keys(SElink_raw* sr, size_t nkeys)
{
	if (nkeys > sr->neg.keys_cap) {
		if (sr->neg.keys == NULL) {
			sr->neg.keys = (SElink_raw_negitem*)malloc(nkeys * sizeof(SElink_raw_negitem));
		}
		else {
			sr->neg.keys = (SElink_raw_negitem*)realloc(sr->neg.keys, nkeys * sizeof(SElink_raw_negitem));
		}
		sr->neg.keys_cap = nkeys;
	}
}
void SElink_raw_reserve_algorithms(SElink_raw* sr, size_t nalgorithms)
{
	if (nalgorithms > sr->neg.algorithms_cap) {
		if (sr->neg.algorithms == NULL) {
			sr->neg.algorithms = (SElink_raw_negitem*)malloc(nalgorithms * sizeof(SElink_raw_negitem));
		}
		else {
			sr->neg.algorithms = (SElink_raw_negitem*)realloc(sr->neg.algorithms, nalgorithms * sizeof(SElink_raw_negitem));
		}
		sr->neg.algorithms_cap = nalgorithms;
	}
}


void SElink_raw_add_key(SElink_raw* sr, uint32_t id, const uint8_t* fingerprint)
{
	size_t i = sr->neg.keys_len;
	SElink_raw_reserve_keys(sr, i + 1);
	sr->neg.keys[i].id = id;
	memcpy(sr->neg.keys[i].fingerprint, fingerprint, SELINK_FINGERPRINT_SIZE);
	sr->neg.keys_len += 1;
}
void SElink_raw_add_algorithm(SElink_raw* sr, uint32_t id, const uint8_t* fingerprint)
{
	size_t i = sr->neg.algorithms_len;
	SElink_raw_reserve_algorithms(sr, i + 1);
	sr->neg.algorithms[i].id = id;
	memcpy(sr->neg.algorithms[i].fingerprint, fingerprint, SELINK_FINGERPRINT_SIZE);
	sr->neg.algorithms_len += 1;
}
void SElink_raw_clear(SElink_raw* sr)
{
	sr->neg.keys_cap = 0;
	sr->neg.keys_len = 0;
	free(sr->neg.keys);
	sr->neg.keys = NULL;

	sr->neg.algorithms_cap = 0;
	sr->neg.algorithms_len = 0;
	free(sr->neg.algorithms);
	sr->neg.algorithms = NULL;
}

void SElink_raw_fingerprint(const uint8_t* salt, size_t len, const uint8_t* data, uint8_t* fingerprint)
{
	PBKDF2HmacSha256(data, len , salt, SELINK_SALT_SIZE, 1, fingerprint, SELINK_FINGERPRINT_SIZE);
}

void SElink_raw_d1_write(SElink_raw* sr, size_t* len, uint8_t* data)
{
	size_t i, k, pos;
	uint16_t u16tmp;

	memcpy(data + SELINK_D1_OFF_MAGIC, selink_magic, sizeof(selink_magic));
	data[SELINK_D1_OFF_VER] = selink_ver;
	data[SELINK_D1_OFF_TYPE] = SELINK_TYPE_D1;

	u16tmp = (uint16_t)(sr->neg.keys_len);
	SE3_SET16(data, SELINK_D1_OFF_NKEYS, u16tmp);
	
	u16tmp = (uint16_t)(sr->neg.algorithms_len);
	SE3_SET16(data, SELINK_D1_OFF_NALGORITHMS, u16tmp);

	memcpy(data + SELINK_D1_OFF_SALT, sr->salt, SELINK_SALT_SIZE);

	k = 0;
	for (i = 0; i < sr->neg.keys_len; i++) {
		pos = SELINK_D1_OFF_FINGERPRINTS + k * SELINK_FINGERPRINT_SIZE;
		memcpy(data + pos, sr->neg.keys[i].fingerprint, SELINK_FINGERPRINT_SIZE);
		k++;
	}
	for (i = 0; i < sr->neg.algorithms_len; i++) {
		pos = SELINK_D1_OFF_FINGERPRINTS + k*SELINK_FINGERPRINT_SIZE;
		memcpy(data + pos, sr->neg.algorithms[i].fingerprint, SELINK_FINGERPRINT_SIZE);
		k++;
	}
	*len = SELINK_D1_OFF_FINGERPRINTS + k*SELINK_FINGERPRINT_SIZE;
}

bool _SElink_raw_check_packet(const uint8_t* data, uint8_t type)
{
	if (memcmp(data + SELINK_D1_OFF_MAGIC, selink_magic, sizeof(selink_magic))) {
		return false;
	}
	if (data[SELINK_D1_OFF_VER] != selink_ver) {
		return false;
	}
	if (data[SELINK_D1_OFF_TYPE] != type) {
		return false;
	}
	return true;
}

bool SElink_raw_d1_get_size(SElink_raw* sr, size_t len, const uint8_t* data, size_t* size)
{
	uint16_t u16tmp = 0;
	size_t nkeys = 0, nalgorithms = 0;
	if (len != SELINK_D1_REQUIRED) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_D1)) {
		return false;
	}
	SE3_GET16(data, SELINK_D1_OFF_NKEYS, u16tmp);
	nkeys = u16tmp;
	SE3_GET16(data, SELINK_D1_OFF_NALGORITHMS, u16tmp);
	nalgorithms = u16tmp;

	*size = _SELINK_D1_SIZE(nkeys, nalgorithms);

	return true;
}


bool SElink_raw_d1_read(SElink_raw* sr, size_t len, const uint8_t* data)
{
	size_t nkeys, nalgorithms;
	size_t expected_len;
	uint16_t u16tmp;
	if (len < SELINK_D1_REQUIRED) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_D1)) {
		return false;
	}
	
	SE3_GET16(data, SELINK_D1_OFF_NKEYS, u16tmp);
	nkeys = u16tmp;
	SE3_GET16(data, SELINK_D1_OFF_NALGORITHMS, u16tmp);
	nalgorithms = u16tmp;
	expected_len = _SELINK_D1_SIZE(nkeys, nalgorithms);
	if (len != expected_len) {
		return false;
	}
	memcpy(sr->salt, data + SELINK_D1_OFF_SALT, SELINK_SALT_SIZE);
	return true;
}

bool SElink_raw_d1_match(SElink_raw* sr, size_t len, const uint8_t* data)
{
	// TODO use maps to reduce complexity

	size_t nkeys, nalgorithms;
	size_t expected_len;
	uint16_t u16tmp;

	size_t i, j, pos;
	bool found;

	sr->choice.key_index = -1;
	sr->choice.algorithm_index = -1;

	if (len < SELINK_D1_REQUIRED) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_D1)) {
		return false;
	}

	SE3_GET16(data, SELINK_D1_OFF_NKEYS, u16tmp);
	nkeys = u16tmp;
	SE3_GET16(data, SELINK_D1_OFF_NALGORITHMS, u16tmp);
	nalgorithms = u16tmp;
	expected_len = _SELINK_D1_SIZE(nkeys, nalgorithms);
	if (len != expected_len) {
		return false;
	}

	found = false;
	for (j = 0; j < sr->neg.keys_len; j++) {
		for (i = 0; i < nkeys; i++) {
			pos = SELINK_D1_OFF_FINGERPRINTS + i * SELINK_FINGERPRINT_SIZE;
			if (!memcmp(data + pos, sr->neg.keys[j].fingerprint, SELINK_FINGERPRINT_SIZE)) {
				found = true;
				break;
			}
		}
		if (found)break;
	}
	if (found) {
		sr->choice.key_index = (int)i;
		memcpy(&(sr->choice.key), &(sr->neg.keys[j]), sizeof(SElink_raw_negitem));
	}

	found = false;
	for (j = 0; j < sr->neg.algorithms_len; j++) {
		for (i = 0; i < nalgorithms; i++) {
			pos = SELINK_D1_OFF_FINGERPRINTS + (nkeys + i)*SELINK_FINGERPRINT_SIZE;
			if (!memcmp(data + pos, sr->neg.algorithms[j].fingerprint, SELINK_FINGERPRINT_SIZE)) {
				found = true;
				break;
			}
		}
		if (found)break;
	}
	if (found) {
		sr->choice.algorithm_index = (int)i;
		memcpy(&(sr->choice.algorithm), &(sr->neg.algorithms[j]), sizeof(SElink_raw_negitem));
	}
	
	return true;
}

void SElink_raw_d2_write(SElink_raw* sr, size_t* len, uint8_t* data)
{
	uint16_t u16tmp;
	memcpy(data + SELINK_D2_OFF_MAGIC, selink_magic, sizeof(selink_magic));
	data[SELINK_D2_OFF_VER] = selink_ver;
	data[SELINK_D2_OFF_TYPE] = SELINK_TYPE_D2;
	if (sr->choice.key_index >= 0) {
		u16tmp = (uint16_t)(sr->choice.key_index);
	}
	else {
		u16tmp = 0xFFFF;
	}
	SE3_SET16(data, SELINK_D2_OFF_KEY, u16tmp);
	if (sr->choice.algorithm_index >= 0) {
		u16tmp = (uint16_t)sr->choice.algorithm_index;
	}
	else {
		u16tmp = 0xFFFF;
	}
	SE3_SET16(data, SELINK_D2_OFF_ALGORITHM, u16tmp);
	memcpy(data + SELINK_D2_OFF_ADD, sr->conf.additional, SELINK_ADD_SIZE);
	*len = SELINK_D2_SIZE;
}

bool SElink_raw_d2_read(SElink_raw* sr, size_t len, const uint8_t* data)
{
	uint16_t u16tmp;

	sr->choice.key_index = -1;
	sr->choice.algorithm_index = -1;

	if (len != SELINK_D2_SIZE) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_D2)) {
		return false;
	}

	SE3_GET16(data, SELINK_D2_OFF_KEY, u16tmp);
	if (u16tmp != 0xFFFF) {
		sr->choice.key_index = u16tmp;
		if (sr->choice.key_index >= sr->neg.keys_len) {
			return false;
		}
		memcpy(&(sr->choice.key), &(sr->neg.keys[sr->choice.key_index]), sizeof(SElink_raw_negitem));
	}
	SE3_GET16(data, SELINK_D2_OFF_ALGORITHM, u16tmp);
	if (u16tmp != 0xFFFF) {
		sr->choice.algorithm_index = u16tmp;
		if (sr->choice.algorithm_index >= sr->neg.algorithms_len) {
			return false;
		}
		memcpy(&(sr->choice.algorithm), &(sr->neg.algorithms[sr->choice.algorithm_index]), sizeof(SElink_raw_negitem));
	}

	memcpy(sr->conf.additional, data + SELINK_D2_OFF_ADD, SELINK_ADD_SIZE);

	return true;
}

void SElink_raw_s1_write(SElink_raw* sr, size_t* len, uint8_t* data)
{
	memcpy(data + SELINK_S1_OFF_MAGIC, selink_magic, sizeof(selink_magic));
	data[SELINK_S1_OFF_VER] = selink_ver;
	data[SELINK_S1_OFF_TYPE] = SELINK_TYPE_S1;
	memcpy(data + SELINK_S1_OFF_ADD, sr->conf.additional, SELINK_ADD_SIZE);
	memcpy(data + SELINK_S1_OFF_SALT, sr->salt, SELINK_SALT_SIZE);
	memcpy(data + SELINK_S1_OFF_FINGERPRINTS + 0 * SELINK_FINGERPRINT_SIZE, sr->choice.key.fingerprint, SELINK_FINGERPRINT_SIZE);
	memcpy(data + SELINK_S1_OFF_FINGERPRINTS + 1 * SELINK_FINGERPRINT_SIZE, sr->choice.algorithm.fingerprint, SELINK_FINGERPRINT_SIZE);

	*len = SELINK_S1_SIZE;
}
bool SElink_raw_s1_read(SElink_raw* sr, size_t len, const uint8_t* data)
{
	if (len != SELINK_S1_SIZE) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_S1)) {
		return false;
	}

	memcpy(sr->salt, data + SELINK_S1_OFF_SALT, SELINK_SALT_SIZE);
	memcpy(sr->conf.additional, data + SELINK_S1_OFF_ADD, SELINK_ADD_SIZE);
	return true;
}

bool SElink_raw_s1_match(SElink_raw* sr, size_t len, const uint8_t* data)
{
	size_t j, pos;
	bool found;

	sr->choice.key_index = -1;
	sr->choice.algorithm_index = -1;

	if (len != SELINK_S1_SIZE) {
		return false;
	}
	if (!_SElink_raw_check_packet(data, SELINK_TYPE_S1)) {
		return false;
	}

	found = false;
	pos = SELINK_S1_OFF_FINGERPRINTS + 0 * SELINK_FINGERPRINT_SIZE;
	for (j = 0; j < sr->neg.keys_len; j++) {
		if (!memcmp(data + pos, sr->neg.keys[j].fingerprint, SELINK_FINGERPRINT_SIZE)) {
			found = true;
			break;
		}
	}
	if (found) {
		sr->choice.key_index = 0;
		memcpy(&(sr->choice.key), &(sr->neg.keys[j]), sizeof(SElink_raw_negitem));
	}

	found = false;
	pos = SELINK_S1_OFF_FINGERPRINTS + 1 * SELINK_FINGERPRINT_SIZE;
	for (j = 0; j < sr->neg.algorithms_len; j++) {
		if (!memcmp(data + pos, sr->neg.algorithms[j].fingerprint, SELINK_FINGERPRINT_SIZE)) {
			found = true;
			break;
		}
	}
	if (found) {
		sr->choice.algorithm_index = 0;
		memcpy(&(sr->choice.algorithm), &(sr->neg.algorithms[j]), sizeof(SElink_raw_negitem));
	}

	return true;
}

size_t SElink_raw_packet_size(size_t data_size)
{
	size_t size = SELINK_H_SIZE + data_size;
	if (0 != size % SELINK_BLOCK_SIZE) {
		size += SELINK_BLOCK_SIZE - (size % SELINK_BLOCK_SIZE);
	}
	size += SELINK_BLOCK_SIZE + SELINK_AUTH_SIZE;
	return size;
}

void SElink_buffer_free(SElink_buffer* buf)
{
	free(buf->data);
	buf->data = NULL;
	buf->cap = 0;
}

void SElink_buffer_init(SElink_buffer * buf)
{
	buf->data = NULL;
	buf->cap = 0;
	buf->len = 0;
}

void SElink_buffer_grow(SElink_buffer* buf, size_t required)
{
	uint32_t len;
	if (buf->cap < required) {
		len = (uint32_t)(required - 1);
		len |= len >> 1;
		len |= len >> 2;
		len |= len >> 4;
		len |= len >> 8;
		len |= len >> 16;
		len++;
		if (buf->data == NULL) {
			buf->data = (uint8_t*)malloc(len);
		}
		else {
			buf->data = realloc(buf->data, len);
		}
		buf->cap = len;
	}
}

void SElink_buffer_ensure(SElink_buffer* buf, size_t newsize)
{
	if (newsize == 0) {
		return;
	}
	if (buf->cap < newsize) {
		if (buf->data == NULL) {
			buf->data = (uint8_t*)malloc(newsize);
		}
		else {
			buf->data = (uint8_t*)realloc(buf->data, newsize);
		}
		buf->cap = newsize;
	}
}

void SElink_raw_h_write(SElink_raw* sr, uint8_t* header, size_t data_size)
{
	uint32_t u32tmp = (uint32_t)data_size;
	SE3_SET32(header, SELINK_H_OFF_LENGTH, u32tmp);
}
void SElink_raw_h_read(SElink_raw* sr, const uint8_t* header, size_t* data_size)
{
	uint32_t u32tmp = 0;
	SE3_GET32(header, SELINK_H_OFF_LENGTH, u32tmp);
	*data_size = (size_t)u32tmp;
}

void SElink_init(SElink* ctx, se3_session* s)
{
	memcpy(&(ctx->s), s, sizeof(se3_session));
	ctx->device_error = SE3_OK;
	ctx->conf.algo_type = SE3_ALGO_AES_HMACSHA256;
	ctx->conf.key_size = 32;
	ctx->conf.algo_mode = SE3_FEEDBACK_CBC;
	SElink_raw_init(&(ctx->raw));
}

uint16_t SElink_raw_secube_import(SElink_raw* sr, se3_session* s, uint16_t key_size, uint16_t algo_type)
{
	uint16_t r, skip, max_keys, count=0;
	size_t i;
	enum {
		NKEYS = 32
	};
	uint32_t now;
	se3_key* keys;
	se3_algo* algorithms;
	uint8_t fingerprint[SELINK_FINGERPRINT_SIZE];

	keys = (se3_key*)malloc(NKEYS * sizeof(se3_key));
	
	r = SE3_OK;
	skip = 0; max_keys = NKEYS <= (SELINK_KEYS_MAX - skip) ? NKEYS : (SELINK_KEYS_MAX - skip);

	now = (uint32_t)time(0);
	while (skip < SELINK_KEYS_MAX) {
		r = L1_key_list(s, skip, max_keys, sr->salt, keys, &count);
		if (r != SE3_OK || count == 0) {
			break;
		}
		SElink_raw_reserve_keys(sr, sr->neg.keys_len + count);
		for (i = 0; i < count; i++) {
			if (keys[i].validity > now + 10) {
				if (keys[i].data_size == key_size) {
					SElink_raw_add_key(sr, keys[i].id, keys[i].fingerprint);
				}
			}
		}
		skip += count;
	}
	free(keys);
	if (r != SE3_OK) {
		SElink_raw_clear(sr);
		return r;
	}

	algorithms = (se3_algo*)calloc(SELINK_ALGORITHMS_MAX, sizeof(se3_algo));
	r = L1_get_algorithms(s, 0, SELINK_ALGORITHMS_MAX, algorithms, &count);
	if (r == SE3_OK) {
		for (i = 0; i < count; i++) {
			if (algorithms[i].type == algo_type) {
				SElink_raw_fingerprint(sr->salt, SE3_CMD1_CRYPTO_ALGOINFO_NAME_SIZE, algorithms[i].name, fingerprint);
				SElink_raw_add_algorithm(sr, (uint32_t)i, fingerprint);
			}
		}
	}
	free(algorithms);
	if (r != SE3_OK) {
		SElink_raw_clear(sr);
		return r;
	}

	return SE3_OK;
}

static uint16_t _SElink_crypto_init(SElink* ctx)
{
	uint16_t r, u16tmp;
	bool encrypt_initialized = false;
	bool decrypt_initialized = false;

	r = L1_crypto_init(&(ctx->s),
		ctx->raw.choice.algorithm.id, SE3_DIR_ENCRYPT | (ctx->conf.algo_mode),
		ctx->raw.choice.key.id, &(ctx->sid_encrypt));
	if (r != SE3_OK) {
		goto fail;
	}
	encrypt_initialized = true;
	r = L1_crypto_update(&(ctx->s), ctx->sid_encrypt, SE3_CRYPTO_FLAG_SETNONCE,
		SELINK_SALT_SIZE, ctx->raw.salt, 0, NULL, &u16tmp, NULL);
	if (r != SE3_OK) {
		goto fail;
	}

	r = L1_crypto_init(&(ctx->s),
		ctx->raw.choice.algorithm.id, SE3_DIR_DECRYPT | (ctx->conf.algo_mode),
		ctx->raw.choice.key.id, &(ctx->sid_decrypt));
	if (r != SE3_OK) {
		goto fail;
	}
	decrypt_initialized = true;
	r = L1_crypto_update(&(ctx->s), ctx->sid_decrypt, SE3_CRYPTO_FLAG_SETNONCE,
		SELINK_SALT_SIZE, ctx->raw.salt, 0, NULL, &u16tmp, NULL);
	if (r != SE3_OK) {
		goto fail;
	}

	return SELINK_OK;

fail:
	ctx->device_error = r;
	if (encrypt_initialized) {
		L1_crypto_update(&(ctx->s), ctx->sid_encrypt, SE3_CRYPTO_FLAG_FINIT,
			0, NULL, 0, NULL, &u16tmp, NULL);
	}
	if (decrypt_initialized) {
		L1_crypto_update(&(ctx->s), ctx->sid_decrypt, SE3_CRYPTO_FLAG_FINIT,
			0, NULL, 0, NULL, &u16tmp, NULL);
	}
	return SELINK_ERR_DEVICE;	
}

uint16_t SElink_duplex_write_request(SElink* ctx, SElink_buffer* buf)
{
	uint16_t r;
	size_t request_len = 0;

	se3c_rand(SELINK_SALT_SIZE, ctx->raw.salt);
	r = SElink_raw_secube_import(&(ctx->raw), &(ctx->s), ctx->conf.key_size, ctx->conf.algo_type);
	if (r != SE3_OK) {
		ctx->device_error = r;
		return SELINK_ERR_DEVICE;
	}

	SElink_buffer_grow(buf, SELINK_D1_SIZE_MAX);
	SElink_raw_d1_write(&(ctx->raw), &request_len, buf->data);
	buf->len = request_len;

	return SELINK_OK;
}

uint16_t SElink_duplex_get_request_size(SElink* ctx, SElink_buffer* buf, size_t* request_len)
{
	if (!SElink_raw_d1_get_size(&(ctx->raw), buf->len, buf->data, request_len)) {
		return SELINK_ERR_DATA;
	}
	return SELINK_OK;
}

uint16_t SElink_duplex_reply(SElink* ctx, SElink_buffer* buf)
{
	uint16_t r;
	size_t response_len = 0;

	if (!SElink_raw_d1_read(&(ctx->raw), buf->len, buf->data)) {
		return SELINK_ERR_DATA;
	}
	
	r = SElink_raw_secube_import(&(ctx->raw), &(ctx->s), ctx->conf.key_size, ctx->conf.algo_type);
	if (r != SE3_OK) {
		ctx->device_error = r;
		return SELINK_ERR_DEVICE;
	}

	if (!SElink_raw_d1_match(&(ctx->raw), buf->len, buf->data)) {
		return SELINK_ERR_DATA;
	}

	if (ctx->raw.choice.key_index < 0) {
		return SELINK_ERR_NOKEY;
	}
	if (ctx->raw.choice.algorithm_index < 0) {
		return SELINK_ERR_NOALGO;
	}

	SElink_buffer_grow(buf, SELINK_D2_SIZE);
	SE3_SET16(ctx->raw.conf.additional, 0, ctx->conf.algo_mode);
	SElink_raw_d2_write(&(ctx->raw), &response_len, buf->data);
	buf->len = response_len;

	r = _SElink_crypto_init(ctx);
	if (r != SELINK_OK) {
		return r;
	}

	return SELINK_OK;
}

uint16_t SElink_duplex_read_response(SElink* ctx, SElink_buffer* buf)
{
	uint16_t r;
	if (!SElink_raw_d2_read(&(ctx->raw), buf->len, buf->data)) {
		return SELINK_ERR_DATA;
	}
	SE3_GET16(ctx->raw.conf.additional, 0, ctx->conf.algo_mode);
	r = _SElink_crypto_init(ctx);
	if (r != SELINK_OK) {
		return r;
	}

	return SELINK_OK;
}

uint16_t SElink_simplex_write_invite(SElink* ctx, SElink_buffer* buf)
{
	size_t invite_len = 0;
	SE3_SET16(ctx->raw.conf.additional, 0, ctx->conf.algo_mode);
	SElink_buffer_grow(buf, SELINK_S1_SIZE);

	SElink_raw_s1_write(&(ctx->raw), &invite_len, buf->data);
	buf->len = invite_len;

	return SELINK_OK;
}



uint16_t SElink_simplex_read_invite(SElink* ctx, SElink_buffer* buf)
{
	uint16_t r;

	if (!SElink_raw_s1_read(&(ctx->raw), buf->len, buf->data)) {
		return SELINK_ERR_DATA;
	}

	r = SElink_raw_secube_import(&(ctx->raw), &(ctx->s), ctx->conf.key_size, ctx->conf.algo_type);
	if (r != SELINK_OK) {
		return r;
	}

	if (!SElink_raw_s1_match(&(ctx->raw), buf->len, buf->data)) {
		return SELINK_ERR_DATA;
	}

	if (ctx->raw.choice.key_index < 0) {
		return SELINK_ERR_NOKEY;
	}
	if (ctx->raw.choice.algorithm_index < 0) {
		return SELINK_ERR_NOALGO;
	}

	r = _SElink_crypto_init(ctx);
	if (r != SELINK_OK) {
		return r;
	}

	return SELINK_OK;
}

uint16_t SElink_write(SElink * ctx, size_t data_len, const uint8_t * data, SElink_buffer* buf)
{
	size_t packet_len, padded_data_len;
	uint16_t dataout_len = 0;
	uint16_t r;

	bool last_crypto = false;
	size_t nchunks, nrem, i;
	enum {
		CHUNK_MAX = SE3_CRYPTO_MAX_DATAIN - SELINK_AUTH_MAX
	};

	// compute total packet size
	padded_data_len = (SELINK_H_SIZE + data_len);
	if (padded_data_len % SELINK_BLOCK_SIZE != 0) {
		padded_data_len += SELINK_BLOCK_SIZE - (padded_data_len % SELINK_BLOCK_SIZE);
	}
	packet_len = SELINK_BLOCK_SIZE + padded_data_len + SELINK_AUTH_SIZE;

	// expand internal buffer
	SElink_buffer_grow(buf, packet_len - SELINK_AUTH_SIZE + SELINK_AUTH_MAX);
	
	// generate random IV
	se3c_rand(SELINK_BLOCK_SIZE, buf->data);

	// write header
	SElink_raw_h_write(&(ctx->raw), buf->data + SELINK_BLOCK_SIZE, data_len);

	// write data
	memcpy(buf->data + SELINK_BLOCK_SIZE + SELINK_H_SIZE, data, data_len);
	// add padding
	if ((SELINK_H_SIZE + data_len) < padded_data_len) {
		memset(buf->data + SELINK_BLOCK_SIZE + SELINK_H_SIZE + data_len, 0, padded_data_len - (SELINK_H_SIZE + data_len));
	}
	buf->len = packet_len;
	
	// set iv
	r = L1_crypto_update(&(ctx->s), ctx->sid_encrypt, SE3_CRYPTO_FLAG_RESET,
		SELINK_BLOCK_SIZE, buf->data, 0, NULL, &dataout_len, NULL);
	if (r != SE3_OK) {
		ctx->device_error = r;
		return SELINK_ERR_DEVICE;
	}

	// encrypt
	nchunks = padded_data_len / CHUNK_MAX;
	nrem = padded_data_len % CHUNK_MAX;
	
	for (i = 0; i < nchunks; i++) {
		last_crypto = ((i == nchunks - 1) && (nrem == 0));

		r = L1_crypto_update(&(ctx->s), ctx->sid_encrypt,
			(last_crypto) ? (SE3_CRYPTO_FLAG_AUTH) : (0) ,
			0, NULL,
			CHUNK_MAX, buf->data + SELINK_BLOCK_SIZE + i*CHUNK_MAX,
			&dataout_len, buf->data + SELINK_BLOCK_SIZE + i*CHUNK_MAX);
		if (r != SE3_OK) {
			ctx->device_error = r;
			return SELINK_ERR_DEVICE;
		}
	}
	if (nrem > 0) {
		r = L1_crypto_update(&(ctx->s), ctx->sid_encrypt,
			SE3_CRYPTO_FLAG_AUTH,
			0, NULL,
			(uint16_t)nrem, buf->data + SELINK_BLOCK_SIZE + nchunks*CHUNK_MAX,
			&dataout_len, buf->data + SELINK_BLOCK_SIZE + nchunks*CHUNK_MAX);
		if (r != SE3_OK) {
			ctx->device_error = r;
			return SELINK_ERR_DEVICE;
		}
	}

	return SELINK_OK;
}

uint16_t SElink_read_header(SElink* ctx, size_t data_len, const uint8_t* data, size_t* remaining_bytes)
{
	uint16_t dataout_len = 0;
	uint16_t r;

	if (data_len != SELINK_HEADER_REQUIRED) {
		return SELINK_ERR_DATA;
	}
	
	r = L1_crypto_update(&(ctx->s), ctx->sid_decrypt,
		SE3_CRYPTO_FLAG_RESET,
		SELINK_BLOCK_SIZE, data,
		SELINK_BLOCK_SIZE, data + SELINK_BLOCK_SIZE,
		&dataout_len, ctx->recv_header.block);
	if (r != SE3_OK) {
		ctx->device_error = r;
		return SELINK_ERR_DEVICE;
	}
	
	SElink_raw_h_read(&(ctx->raw), ctx->recv_header.block, &(ctx->recv_header.len));

	ctx->recv_header.padded_data_len = ctx->recv_header.len + SELINK_H_SIZE;
	if (ctx->recv_header.padded_data_len % SELINK_BLOCK_SIZE != 0) {
		ctx->recv_header.padded_data_len += SELINK_BLOCK_SIZE - (ctx->recv_header.padded_data_len % SELINK_BLOCK_SIZE);
	}
	ctx->recv_header.packet_len = SELINK_BLOCK_SIZE + ctx->recv_header.padded_data_len + SELINK_AUTH_SIZE;
	*remaining_bytes = ctx->recv_header.packet_len - SELINK_HEADER_REQUIRED;
	
	return SELINK_OK;
}

uint16_t SElink_read(SElink* ctx, size_t data_len, const uint8_t* data, SElink_buffer* buf)
{
	size_t remaining;

	bool last_crypto = false;
	size_t nchunks, nrem, i;
	enum {
		CHUNK_MAX = SE3_CRYPTO_MAX_DATAIN - B5_SHA256_DIGEST_SIZE
	};
	uint16_t r, dataout_len;

	remaining = ctx->recv_header.packet_len - SELINK_HEADER_REQUIRED;
	if (data_len != remaining) {
		return SELINK_ERR_DATA;
	}

	SElink_buffer_grow(buf, ctx->recv_header.packet_len - SELINK_AUTH_SIZE + SELINK_AUTH_MAX);

	memcpy(buf->data, ctx->recv_header.block + SELINK_H_SIZE, SELINK_BLOCK_SIZE - SELINK_H_SIZE);
	buf->len = SELINK_BLOCK_SIZE - SELINK_H_SIZE;

	// decrypt
	nchunks = (remaining - SELINK_AUTH_SIZE) / CHUNK_MAX;
	nrem = (remaining - SELINK_AUTH_SIZE) % CHUNK_MAX;

	for (i = 0; i < nchunks; i++) {
		last_crypto = ((i == nchunks - 1) && (nrem == 0));

		r = L1_crypto_update(&(ctx->s), ctx->sid_decrypt,
			(last_crypto) ? (SE3_CRYPTO_FLAG_AUTH) : (0),
			0, NULL,
			CHUNK_MAX, (uint8_t*)data + i*CHUNK_MAX,
			&dataout_len, buf->data + buf->len);
		if (r != SE3_OK) {
			ctx->device_error = r;
			return SELINK_ERR_DEVICE;
		}
		buf->len += CHUNK_MAX;
	}
	if (nrem > 0) {
		r = L1_crypto_update(&(ctx->s), ctx->sid_decrypt,
			SE3_CRYPTO_FLAG_AUTH,
			0, NULL,
			(uint16_t)nrem, (uint8_t*)data + nchunks*CHUNK_MAX,
			&dataout_len, buf->data + buf->len);
		if (r != SE3_OK) {
			ctx->device_error = r;
			return SELINK_ERR_DEVICE;
		}
		buf->len += nrem;
	}

	if (memcmp(data + remaining - SELINK_AUTH_SIZE, buf->data + buf->len, SELINK_AUTH_SIZE)) {
		return SELINK_ERR_AUTH;
	}

	buf->len = ctx->recv_header.len;

	return SELINK_OK;
}

void SElink_destroy(SElink* ctx)
{
	uint16_t u16tmp;
	SElink_raw_clear(&(ctx->raw));

	L1_crypto_update(&(ctx->s), ctx->sid_encrypt, SE3_CRYPTO_FLAG_FINIT,
		0, NULL, 0, NULL, &u16tmp, NULL);
	L1_crypto_update(&(ctx->s), ctx->sid_decrypt, SE3_CRYPTO_FLAG_FINIT,
		0, NULL, 0, NULL, &u16tmp, NULL);
}

