#pragma once
#include "secube/L1.h"
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/** SElink API error definitions */
enum {
	SELINK_OK = 0,  ///< the operation succeeded
	SELINK_ERR_DATA = 100,  ///< packet data failed to validate
	SELINK_ERR_AUTH = 101,  ///< received data failed to authenticate
	SELINK_ERR_NOKEY = 200,  ///< no matching key found
	SELINK_ERR_NOALGO = 201,  ///< no matching algorithm found
	SELINK_ERR_DEVICE = 1000  ///< a SEcube operation failed
};

/** SElink API structure size definitions */
enum {
	SELINK_SALT_SIZE = SE3_KEY_SALT_SIZE,
	SELINK_AUTH_SIZE = 16,
	SELINK_AUTH_MAX = 64,
	SELINK_BLOCK_SIZE = 16,
	SELINK_FINGERPRINT_SIZE = SE3_KEY_FINGERPRINT_SIZE,
	SELINK_ADD_SIZE = 16,
	SELINK_HEADER_REQUIRED = 2 * SELINK_BLOCK_SIZE,

	SELINK_KEYS_MAX = 512,
	SELINK_ALGORITHMS_MAX = 64,

	SELINK_DATA_MAX = 7000  // TODO
};

/** SElink D1 packet definitions */
enum {
	SELINK_D1_OFF_MAGIC = 0,
	SELINK_D1_OFF_VER = 2,
	SELINK_D1_OFF_TYPE = 3,
	SELINK_D1_OFF_NKEYS = 4,
	SELINK_D1_OFF_NALGORITHMS = 6,
	SELINK_D1_OFF_SALT = 8,
	SELINK_D1_OFF_FINGERPRINTS = SELINK_D1_OFF_SALT + SELINK_SALT_SIZE,

	SELINK_D1_REQUIRED = 8
};
#define _SELINK_D1_SIZE(keys, algorithms) (SELINK_D1_OFF_FINGERPRINTS + ((keys) + (algorithms)) * SELINK_FINGERPRINT_SIZE)
enum {
	SELINK_D1_SIZE_MAX = _SELINK_D1_SIZE(SELINK_KEYS_MAX, SELINK_ALGORITHMS_MAX)
};

/** SElink D2 packet definitions */
enum {
	SELINK_D2_OFF_MAGIC = 0,
	SELINK_D2_OFF_VER = 2,
	SELINK_D2_OFF_TYPE = 3,
	SELINK_D2_OFF_KEY = 4,
	SELINK_D2_OFF_ALGORITHM = 6,
	SELINK_D2_OFF_ADD = 8,

	SELINK_D2_SIZE = SELINK_D2_OFF_ADD + SELINK_ADD_SIZE
};

/** SElink S1 packet definitions */
enum {
	SELINK_S1_OFF_MAGIC = 0,
	SELINK_S1_OFF_VER = 2,
	SELINK_S1_OFF_TYPE = 3,
	SELINK_S1_OFF_ADD = 4,
	SELINK_S1_OFF_SALT = 4 + SELINK_ADD_SIZE,
	SELINK_S1_OFF_FINGERPRINTS = SELINK_S1_OFF_SALT + SELINK_SALT_SIZE,

	SELINK_S1_SIZE = SELINK_S1_OFF_FINGERPRINTS + 2 * SELINK_FINGERPRINT_SIZE
};

/** SElink header of data packet definitions */
enum {
	SELINK_H_OFF_LENGTH = 0,
	SELINK_H_SIZE = 4
};


/**
* \defgroup SElink buffer
* @{
*/

/** \brief SElink resizeable buffer object */
typedef struct SElink_buffer_ {
	uint8_t* data;  ///< buffer data
	size_t cap;  ///< maximum capacity
	size_t len;  ///< actual length of store data
} SElink_buffer;

/** \brief Initialize a SElink buffer
 *  
 *  Initialize a buffer object to an empty buffer.
 *  \param SElink buffer object
 */
void SElink_buffer_init(SElink_buffer* buf);

/** \brief Ensure capacity for SElink buffer
 *  
 *  Expand buffer capacity if currently lesser than the desired capacity.
 *  \param SElink buffer object
 */
void SElink_buffer_ensure(SElink_buffer* buf, size_t newsize);

/** \brief Ensure capacity for SElink buffer
 *
 *  Expand buffer capacity if currently lesser than the desired capacity,
 *    to the nearest power of 2 greater or equal to the desired capacity.
 *  \param SElink buffer object
 */
void SElink_buffer_grow(SElink_buffer* buf, size_t required);

/** \brief Dispose SElink buffer
 *
 *  Free allocated memory, reset size and capacity to zero.
 *  \param SElink buffer object
 */
void SElink_buffer_free(SElink_buffer* buf);

/**@}*/


/**
* \defgroup SElink raw API
* @{
*/

 /** \brief single item to be negotiated
 *
 *  Holds id and fingerprint of a key or an algorithm
 */
typedef struct SElink_raw_negitem_ {
	uint32_t id;
	uint8_t fingerprint[SELINK_FINGERPRINT_SIZE];
} SElink_raw_negitem;

/** \brief SElink raw interface context */
typedef struct SElink_raw_ {
	uint8_t salt[SELINK_SALT_SIZE];  /// salt for fingerprints and session keys
	struct {
		SElink_raw_negitem* keys;  ///< available keys
		size_t keys_cap;  ///< maximum capacity of the keys buffer
		size_t keys_len;  ///< number of available keys
		SElink_raw_negitem* algorithms;  ///< available algorithms
		size_t algorithms_cap;  ///< maximum capacity of the algorithms buffer
		size_t algorithms_len;  ///< number of available algorithms
	} neg;
	struct {
		SElink_raw_negitem key;  ///< chosen key
		int key_index;  ///< index of chosen key
		SElink_raw_negitem algorithm;  ///< chosen algorithm
		int algorithm_index;  ///< index of chosen algorithm
	} choice;
	struct {
		uint8_t additional[SELINK_ADD_SIZE];  ///< additional configuration data
	} conf;
} SElink_raw;

/** \brief Initialize SElink raw context
 * 
 *  \param sr SElink raw context
 */
void SElink_raw_init(SElink_raw* sr);

/** \brief Reserve capacity for keys
 * 
 *  Reserve capacity for the keys buffer. If the number of keys
 *    to be added is known, this prevents frequent reallocation of the keys buffer.
 *  \param sr SElink raw context
 *  \param nkeys new maximum capacity
 */
void SElink_raw_reserve_keys(SElink_raw* sr, size_t nkeys);

/** \brief Reserve capacity for algorithms
 * 
 *  Reserve capacity for the algorithms buffer. If the number of keys
 *  to be added is known, this prevents frequent reallocation of the keys buffer.
 *  \param sr SElink raw context
 *  \param nalgorithms new maximum capacity
 */
void SElink_raw_reserve_algorithms(SElink_raw* sr, size_t nalgorithms);

/** \brief add key to context
 *  
 *  Add an available key to the keys buffer in the SElink raw context
 *  \param sr SElink raw context
 *  \param id key id
 *  \params fingerprint key fingerprint
 */
void SElink_raw_add_key(SElink_raw* sr, uint32_t id, const uint8_t* fingerprint);

/** \brief add algorithm to context
 *
 *  Add an available algorithm to the algorithms buffer in the SElink raw context
 *  \param sr SElink raw context
 *  \param id algorithm id
 *  \params fingerprint algorithm fingerprint
 */
void SElink_raw_add_algorithm(SElink_raw* sr, uint32_t id, const uint8_t* fingerprint);

/** \brief Dispose keys and algorithms buffers
 *
 *  Dispose keys and algorithms buffers, freeing associated memory. This operation can be safely
 *    performed after a duplex negotiation is concluded. 
 *  \param sr SElink raw context
 */
void SElink_raw_clear(SElink_raw* sr);

/** \brief Write D1 packet
 *
 *  Generate the D1 packet 
 *  \param sr SElink raw context
 *  \param[out] len length of packet data. max SELINK_D1_SIZE_MAX
 *  \param[out] data packet data
 */
void SElink_raw_d1_write(SElink_raw* sr, size_t* len, uint8_t* data);

/** \brief Get D1 packet size
 *
 *  Read the D1 packet header and get the full packet size
 *  \param sr SElink raw context
 *  \param len length of input data. Must be SELINK_D1_REQUIRED
 *  \param data input data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_d1_get_size(SElink_raw* sr, size_t len, const uint8_t* data, size_t* size);

/** \brief Read D1 packet
 *  
 *  Validate a D1 packet, get the salt and store it into the context.
 *  \param sr SElink raw context
 *  \param len length of packet data
 *  \param data packet data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_d1_read(SElink_raw* sr, size_t len, const uint8_t* data);

/** \brief Find match in D1 packet
 *  
 *  Find matching key and algorithm and store them into the context.
 *  \param sr SElink raw context
 *  \param len length of packet data
 *  \param data packet data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_d1_match(SElink_raw* sr, size_t len, const uint8_t* data);

/** \brief Write D2 packet
 *  
 *  Write the chosen key and algorithm, plus additional configuration, to a D2 packet.
 *  \param sr SElink raw context
 *  \param[out] len length of packet data. always SELINK_D2_SIZE
 *  \param[out] packet data
 */
void SElink_raw_d2_write(SElink_raw* sr, size_t* len, uint8_t* data);

/** \brief Read D2 packet
 *  
 *  Validate and read a D2 packet and store the chosen key and algorithm into the context.
 *  \param sr SElink raw context
 *  \param len length of packet data. Must be SELINK_D2_SIZE
 *  \param data packet data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_d2_read(SElink_raw* sr, size_t len, const uint8_t* data);

/** \brief Write S1 packet
 *  
 *  Write the chosen key and algorithm, plus additional configuration, to a S1 packet
 *  \param sr SElink raw context
 *  \param[out] len length of packet data. Always SELINK_S1_SIZE
 *  \param[out] data packet data
 */
void SElink_raw_s1_write(SElink_raw* sr, size_t* len, uint8_t* data);

/** \brief Read S1 packet
 *  
 *  Validate a S1 packet, get the salt and store it into the context.
 *  \param sr SElink raw context
 *  \param len length of packet data. Must be SELINK_S1_SIZE
 *  \param data packet data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_s1_read(SElink_raw* sr, size_t len, const uint8_t* data);

/** \brief Find match in S1 packet
 *  
 *  Find matching key and algorithm and store them into the context.
 *  \param sr SElink raw context
 *  \param len length of packet data
 *  \param data packet data
 *  \return false if validation of data fails, otherwise true
 */
bool SElink_raw_s1_match(SElink_raw* sr, size_t len, const uint8_t* data);

/** \brief Write data packet header
 *  
 *  \param sr SElink raw context
 *  \param[out] header header data
 *  \param data_size size of message
 */
void SElink_raw_h_write(SElink_raw* sr, uint8_t* header, size_t data_size);

/** \brief Read data packet header
 *  
 *  \param sr SElink raw context
 *  \param header header data
 *  \param[out] data_size size of message
 */
void SElink_raw_h_read(SElink_raw* sr, const uint8_t* header, size_t* data_size);

/** \brief Generate fingerprint
 *  
 *  Generate a fingerprint from some data
 *  \param salt 32-byte salt for the fingerprint
 *  \param len length of input data
 *  \param data input data
 *  \param[out] fingerprint 32-byte fingerprint data
 */
void SElink_raw_fingerprint(const uint8_t* salt, size_t len, const uint8_t* data, uint8_t* fingerprint);

/** \brief Import keys and algorithms from SEcube
 *  
 *  Import all keys and algorithms from a SEcube device, and add them to the context.
 *  \param sr SElink raw context
 *  \param s SEcube session
 *  \param key_size allowed key size. Different sizes are ignored
 *  \param algo_type allowed alogorithm type. Different types are ignored
 *  \return SE3_OK on success, otherwise the error code received from the last failed SEcube operation
 */
uint16_t SElink_raw_secube_import(SElink_raw* sr, se3_session* s, uint16_t key_size, uint16_t algo_type);

/** \brief Get expected packet size
 *  
 *  Get the expected encrypted packet size from the size of clear data,
 *   taking into account the IV, header, padding and authentication overhead.
 *  \param data_size size of clear data
 *  \return size of encrypted packet
 */
size_t SElink_raw_packet_size(size_t data_size);

/**@}*/



/**
 * \defgroup SElink API
 * @{
 */

/** \brief SElink interface context */
typedef struct SElink_ {
	se3_session s;  ///< SEcube session
	uint32_t sid_encrypt;  ///< crypto session id for encryption
	uint32_t sid_decrypt;  ///< crypto session id for decryption
	SElink_raw raw;  ///< SElink raw context
	struct {
		size_t len;  ///< actual length of data
		size_t packet_len;  ///< total length of packet
		size_t padded_data_len;  ///< length of encrypted data
		uint8_t block[SELINK_BLOCK_SIZE];  ///< first decrypted block
	} recv_header;  ///< decrypted data from the last received header
	struct {
		uint16_t key_size;  ///< allowed key size
		uint16_t algo_type;  ///< allowed algorithm type
		uint16_t algo_mode;  ///< algorithm mode
	} conf;  ///< SElink configuration
	uint16_t device_error;  ///< last SEcube device error code
} SElink;

/** \brief Initialize SElink context
 * 
 *  \param ctx SElink context
 */
void SElink_init(SElink* ctx, se3_session* s);

/** \brief Destroy SElink context
 *  
 *  Destroy a SElink context and free all resources.
 *  \param ctx SElink context
 */
void SElink_destroy(SElink* ctx);

/** \brief Write duplex request
 *  
 *  Write a duplex request packet. Send this packet to a SElink peer to
 *    negotiate the connection parameters.
 *  \param ctx SElink context
 *  \param[out] buf buffer to store the packet data
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DEVICE a SEcube operation failed
 */
uint16_t SElink_duplex_write_request(SElink* ctx, SElink_buffer* buf);

/** \brief Get duplex request packet length
 *  
 *  Read a duplex request packet's header to get its total length.
 *  \param ctx SElink context
 *  \param buf buffer containing the request packet's header. Length must be SELINK_D1_REQUIRED
 *  \param[out] request_len full request packet size
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DATA packet data failed to validate
 */
uint16_t SElink_duplex_get_request_size(SElink* ctx, SElink_buffer* buf, size_t* request_len);

/** \brief Reply to duplex request
 *  
 *  Read an incoming duplex request packet and write a response packet.
 *  Send the response packet to the requesting peer to finalize the
 *  negotiation of connection parameters.
 *  \param ctx SElink context
 *  \param[in,out] buf buffer containing the request packet data, also where the 
 *    response packet data will be stored.
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DEVICE a SEcube operation failed
 *  \retval SELINK_ERR_DATA packet data failed to validate
 *  \retval SELINK_ERR_NOKEY no matching key found
 *  \retval SELINK_ERR_NOALGO no matching algorithm found
 */
uint16_t SElink_duplex_reply(SElink* ctx, SElink_buffer* buf);

/** \brief Read duplex response
 *  
 *  Read a duplex response packet and store the negotiated parameters into the context.
 *  \param ctx SElink context
 *  \param buf buffer containing the response packet data
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DATA packet data failed to validate
 */
uint16_t SElink_duplex_read_response(SElink* ctx, SElink_buffer* buf);

/** \brief Write simplex invite
 *  
 *  Write a simplex invite packet. Send this packet to a SElink peer to invite him to join a
 *    SElink network, using the previously negotiated parameters.
 *  \param ctx SElink context
 *  \param[out] buf buffer to store the packet data. Length is always SELINK_S1_SIZE
 *  \retval SELINK_OK the operation succeeded
 */
uint16_t SElink_simplex_write_invite(SElink* ctx, SElink_buffer* buf);

/** \brief Read simplex invite
 *  
 *  Read a simplex invite packet, and adopt the specified key and algorithm if available.
 *  \param ctx SElink context
 *  \param buf buffer containing the invite packet data. Length must be SELINK_S1_SIZE
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DEVICE a SEcube operation failed
 *  \retval SELINK_ERR_DATA packet data failed to validate
 *  \retval SELINK_ERR_NOKEY no matching key found
 *  \retval SELINK_ERR_NOALGO no matching algorithm found
 */
uint16_t SElink_simplex_read_invite(SElink* ctx, SElink_buffer* buf);

/** \brief Write data packet
 *  
 *  Encrypt data and generate a data packet
 *  \param ctx SElink context
 *  \param data_len length of data to be sent
 *  \param data data to be sent
 *  \param[out] buf buffer to store the packet data
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DEVICE a SEcube operation failed
 */
uint16_t SElink_write(SElink* ctx, size_t data_len, const uint8_t* data, SElink_buffer* buf);

/** \brief Read data packet header
 *  
 *  Read and decrypt a data packet's header to determine the packet's size.
 *  \param ctx SElink context
 *  \param data_len length of header data. Must be SELINK_HEADER_REQUIRED
 *  \param data header data
 *  \param[out] remaining_bytes number of bytes needed to complete the packet
 *  \retval SELINK_OK the operation succeeded
 *  \retval SELINK_ERR_DEVICE a SEcube operation failed
 */
uint16_t SElink_read_header(SElink* ctx, size_t data_len, const uint8_t* data, size_t* remaining_bytes);

/** \brief Read data packet
 *  
 *  Read data packet and decrypt data
 *  \param ctx SElink context
 *  \param data_len length of received data (no header)
 *  \param data received data (no header)
 *  \param[out] buf buffer to store the received data, decrypted
 *  \retval SELINK_ERR_AUTH received data failed to authenticate
 */
uint16_t SElink_read(SElink* ctx, size_t data_len, const uint8_t* data, SElink_buffer* buf);

/**@}*/

#ifdef __cplusplus
}
#endif
