#include <iostream>
#include <cassert>
#include <string>
#include <vector>
#include <array>

#include <selink/SElink.h>

#include <boost/filesystem.hpp>
#include <proxy/provider/provider_soft.h>
#include <proxy/provider/provider_secube.h>
#include <util/rng.h>
#include <util/filesystem.h>

static const std::array<uint8_t, 32> TEST_SERIAL = { 4, 3, 2, 1, 4, 3, 2, 1, 4, 3, 2, 1, 4, 3, 2, 1, 8, 7, 6, 5, 8, 7, 6, 5, 8, 7, 6, 5, 8, 7, 6, 5 };
static const std::vector<uint8_t> TEST_PASSWORD = { 'c','i','a','o' };
static const int TEST_KEY = 1;
static const int TEST_ALGO = SE3_ALGO_AES_HMACSHA256;
static const std::array<uint8_t, 32> TEST_NONCE = { 101, 211, 105, 12, 2, 205, 63, 224, 180, 117, 140, 101, 199, 166, 33, 158, 60, 107, 87, 119, 105, 89, 147, 147, 46, 36, 33, 13, 47, 92, 55, 68 };
static const std::array<uint8_t, 16> TEST_IV = { 194, 228, 108, 226, 113, 168, 48, 214, 95, 44, 116, 51, 207, 241, 153, 143 };


void test_selink_raw()
{
	size_t i;
	uint8_t fingerprint[SELINK_FINGERPRINT_SIZE];
	uint32_t id;

	SElink_raw a, b, c;
	
	uint8_t buf[SELINK_D1_SIZE_MAX];
	size_t len;

	// initialize A
	SElink_raw_init(&a);
	se3c_rand(SELINK_SALT_SIZE, a.salt);
	
	
	// initialize B
	SElink_raw_init(&b);
	se3c_rand(SELINK_ADD_SIZE, b.conf.additional);

	// initialize C
	SElink_raw_init(&c);

	// A, B, C keys
	for (i = 0; i < 16; i++) {
		id = (uint32_t)i + 100;

		se3c_rand(SELINK_FINGERPRINT_SIZE, fingerprint);
		SElink_raw_add_key(&b, id, fingerprint);

		se3c_rand(SELINK_FINGERPRINT_SIZE, fingerprint);
		SElink_raw_add_key(&c, id, fingerprint);

		se3c_rand(SELINK_FINGERPRINT_SIZE, fingerprint);
		SElink_raw_add_key(&a, id, fingerprint);
	}

	// add last A key to B, C
	SElink_raw_add_key(&b, id, fingerprint);
	SElink_raw_add_key(&c, id, fingerprint);

	// A, B, C algorithms
	for (i = 0; i < 6; i++) {
		id = (uint32_t)i;
		se3c_rand(SELINK_FINGERPRINT_SIZE, fingerprint);
		SElink_raw_add_algorithm(&a, id, fingerprint);
		SElink_raw_add_algorithm(&b, id, fingerprint);
		SElink_raw_add_algorithm(&c, id, fingerprint);
	}

	bool rb;

	// A negotiates connection with B (duplex)
	
	// A : write duplex request D1
	SElink_raw_d1_write(&a, &len, buf);

	// B : read duplex request D1
	size_t d1_size;
	assert(len >= SELINK_D1_REQUIRED);
	rb = SElink_raw_d1_get_size(&b, SELINK_D1_REQUIRED, buf, &d1_size);
	assert(rb);
	assert(d1_size == len);
	rb = SElink_raw_d1_read(&b, len, buf);
	assert(rb);
	// B : find matching key and algorithm
	rb = SElink_raw_d1_match(&b, len, buf);
	assert(rb);
	// B : write duplex response D2
	SElink_raw_d2_write(&b, &len, buf);

	// A : read duplex response D2
	rb = SElink_raw_d2_read(&a, len, buf);
	assert(rb);

	// check negotiated parameters
	assert(!memcmp(a.salt, b.salt, SELINK_SALT_SIZE));
	assert(a.choice.algorithm_index != -1);
	assert(a.choice.key_index != -1);
	assert(!memcmp(&a.choice.key.fingerprint, &b.choice.key.fingerprint, SELINK_FINGERPRINT_SIZE));
	assert(!memcmp(&a.choice.algorithm.fingerprint, &b.choice.algorithm.fingerprint, SELINK_FINGERPRINT_SIZE));
	assert(!memcmp(a.conf.additional, b.conf.additional, SELINK_ADD_SIZE));

	// A invites C to join (simplex)

	// A : write simplex invite S1
	SElink_raw_s1_write(&a, &len, buf);

	// C : read simplex invite S1
	SElink_raw_s1_read(&c, len, buf);
	// C : find matching key and algorithm
	SElink_raw_s1_match(&c, len, buf);

	// check negotiated parameters
	assert(!memcmp(a.salt, c.salt, SELINK_SALT_SIZE));
	assert(c.choice.algorithm_index != -1);
	assert(c.choice.key_index != -1);
	assert(!memcmp(&a.choice.key.fingerprint, &c.choice.key.fingerprint, SELINK_FINGERPRINT_SIZE));
	assert(!memcmp(&a.choice.algorithm.fingerprint, &c.choice.algorithm.fingerprint, SELINK_FINGERPRINT_SIZE));
	assert(!memcmp(a.conf.additional, c.conf.additional, SELINK_ADD_SIZE));

}


void test_selink()
{
	se3_disco_it it;
	se3_device dev;
	se3_session s;
	uint16_t r;
	uint8_t pin[32];

	enum {
		TEST_SIZE = 50
	};
	uint8_t test[TEST_SIZE];
	se3c_rand(TEST_SIZE, test);

	memset(pin, 0, sizeof(pin));
	memcpy((char*)pin, "ciao", 4);

	L0_discover_init(&it);
	if (!L0_discover_next(&it)) {
		std::cout << "No device found" << std::endl;
	}
	r = L0_open(&dev, &it.device_info, 1000);
	assert(r == SE3_OK);
	r = L1_login(&s, &dev, pin, SE3_ACCESS_USER);
	assert(r == SE3_OK);
	r = L1_crypto_set_time(&s, (uint32_t)time(0));
	assert(r == SE3_OK);

	SElink a, b, c;
	SElink_buffer buf, buf2;
	size_t needed;

	SElink_init(&a, &s);
	SElink_init(&b, &s);
	b.conf.algo_mode = SE3_FEEDBACK_CBC;

	SElink_init(&c, &s);
	SElink_buffer_init(&buf);
	SElink_buffer_init(&buf2);

	// negotiate connection between A and B (duplex)
	r = SElink_duplex_write_request(&a, &buf);
	assert(r == SELINK_OK);

	size_t tmp = buf.len;
	buf.len = SELINK_D1_REQUIRED;
	
	r = SElink_duplex_get_request_size(&b, &buf, &buf.len);
	assert(r == SELINK_OK);
	assert(tmp == buf.len);

	r = SElink_duplex_reply(&b, &buf);
	assert(r == SELINK_OK);

	r = SElink_duplex_read_response(&a, &buf);
	assert(r == SELINK_OK);

	// A invites C to join (simplex)
	r = SElink_simplex_write_invite(&a, &buf);
	assert(r == SELINK_OK);
	r = SElink_simplex_read_invite(&c, &buf);
	assert(r == SELINK_OK);

	// A sends data to B and C
	r = SElink_write(&a, TEST_SIZE, test, &buf);
	assert(r == SELINK_OK);

	r = SElink_read_header(&b, SELINK_HEADER_REQUIRED, buf.data, &needed);
	assert(r == SELINK_OK);
	assert(buf.len == needed + SELINK_HEADER_REQUIRED);
	r = SElink_read(&b, needed, buf.data + SELINK_HEADER_REQUIRED, &buf2);
	assert(r == SELINK_OK);
	assert(buf2.len == TEST_SIZE);
	assert(!memcmp(buf2.data, test, TEST_SIZE));

	r = SElink_read_header(&c, SELINK_HEADER_REQUIRED, buf.data, &needed);
	assert(r == SELINK_OK);
	assert(buf.len == needed + SELINK_HEADER_REQUIRED);
	r = SElink_read(&c, needed, buf.data + SELINK_HEADER_REQUIRED, &buf2);
	assert(r == SELINK_OK);
	assert(buf2.len == TEST_SIZE);
	assert(!memcmp(buf2.data, test, TEST_SIZE));

	SElink_destroy(&a);
	SElink_destroy(&b);
	SElink_destroy(&c);
	SElink_buffer_free(&buf);
	SElink_buffer_free(&buf2);

	L1_logout(&s);
	L0_close(&s.device);

}


void test_provider(selink::provider::provider& provider, const std::vector<uint8_t>& testdata, std::vector<uint8_t>& enc, std::vector<uint8_t>& dec)
{
	provider.config(TEST_SERIAL, TEST_PASSWORD);
	provider.login();
	
	selink::provider::session::error e;
	std::shared_ptr<selink::provider::session> session;

	// encrypt and authenticate
	session = provider.create_session(TEST_KEY, TEST_ALGO, &TEST_NONCE[0], selink::provider::session::direction::encrypt);
	assert(session);
	enc.resize(testdata.size() + SELINK_AUTH_MAX);
	e = session->update(
		true, &TEST_IV[0],
		true, testdata.size() / selink::provider::session::block_size, &testdata[0], &enc[0],
		true);
	assert(e == selink::provider::session::error::ok);
	enc.resize(testdata.size() + selink::provider::session::auth_size);
	session.reset();

	// decrypt and authenticate
	session = provider.create_session(TEST_KEY, TEST_ALGO, &TEST_NONCE[0], selink::provider::session::direction::decrypt);
	assert(session);
	dec.resize(testdata.size() + SELINK_AUTH_MAX);
	e = session->update(
		true, &TEST_IV[0],
		true, testdata.size() / selink::provider::session::block_size, &testdata[0], &dec[0],
		true);
	assert(e == selink::provider::session::error::ok);
	dec.resize(testdata.size() + selink::provider::session::auth_size);
	session.reset();

}


void test_providers()
{
	selink::provider::provider_soft soft;
	selink::provider::provider_secube secube;
	std::vector<uint8_t> testdata(32);
	selink::util::rng::rand_bytes(&testdata[0], testdata.size());

	std::vector<uint8_t> enc[2], dec[2];

	auto path = selink::util::filesystem::get_exe_path().parent_path() / boost::filesystem::path("keys.json");
	soft.set_keys(selink::provider::provider_soft::keys_from_file(path));

	test_provider(soft, testdata, enc[0], dec[0]);

	test_provider(secube, testdata, enc[1], dec[1]);

	secube.reset();

	assert(enc[0] == enc[1]);
	assert(dec[0] == dec[1]);
}


#define TEST_SELINK_RAW 0
#define TEST_SELINK	0
#define TEST_PROVIDERS 1

int main()
{
	std::cout << "HAI" << std::endl;

	if (TEST_SELINK_RAW) {
		std::cout << "Testing SElink-Raw" << std::endl;
		test_selink_raw();
		std::cout << "PASS" << std::endl << std::endl;
	}

	if (TEST_SELINK) {
		std::cout << "Testing SElink" << std::endl;
		test_selink();
		std::cout << "PASS" << std::endl << std::endl;
	}
	
	if (TEST_PROVIDERS) {
		std::cout << "Testing Providers" << std::endl;
		test_providers();
		std::cout << "PASS" << std::endl << std::endl;
	}

	std::cout << "KTHXBYE" << std::endl;

    return 0;
}

